@@ -1,6 +1,9 @@
+/*	$OpenBSD: displayq.c,v 1.26 2003/06/02 23:36:53 millert Exp $	*/
+/*	$NetBSD: displayq.c,v 1.21 2001/08/30 00:51:50 itojun Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -10,11 +13,7 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. All advertising materials mentioning features or use of this software
- *    must display the following acknowledgement:
- *	This product includes software developed by the University of
- *	California, Berkeley and its contributors.
- * 4. Neither the name of the University nor the names of its contributors
+ * 3. Neither the name of the University nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  *
@@ -32,17 +31,36 @@
  */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)displayq.c	5.13 (Berkeley) 6/1/90";*/
-static char rcsid[] = "$Id: displayq.c,v 1.5 1999/10/14 04:31:43 notting Exp $";
+#if 0
+static const char sccsid[] = "@(#)displayq.c	8.4 (Berkeley) 4/28/95";
+#else
+static const char rcsid[] = "$OpenBSD: displayq.c,v 1.26 2003/06/02 23:36:53 millert Exp $";
+#endif
 #endif /* not lint */
 
+#include <sys/param.h>
+#include <sys/file.h>
+#include <sys/ioctl.h>
+#include <sys/stat.h>
+
+#include <ctype.h>
+#include <errno.h>
+#include <dirent.h>
+#include <fcntl.h>
+#include <signal.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <unistd.h>
+#include <vis.h>
+
+#include "lp.h"
+#include "lp.local.h"
+#include "pathnames.h"
+
 /*
  * Routines to display the state of the queue.
  */
-
-#include "lp.h"
-#include "pathnames.h"
-
 #define JOBCOL	40		/* column for job # in -l format */
 #define OWNCOL	7		/* start of Owner column in normal */
 #define SIZCOL	62		/* start of Size column in normal */
@@ -50,122 +68,159 @@ static char rcsid[] = "$Id: displayq.c,v 1.5 1999/10/14 04:31:43 notting Exp $";
 /*
  * Stuff for handling job specifications
  */
-extern char	*user[];	/* users to process */
-extern int	users;		/* # of users in user array */
 extern int	requ[];		/* job number of spool entries */
 extern int	requests;	/* # of spool requests */
+extern char    *user[];	        /* users to process */
+extern int	users;		/* # of users in user array */
 
-int	lflag;		/* long output option */
-char	current[FILENAMELEN];	/* current file being printed */
-int	garbage;	/* # of garbage cf files */
-int	rank;		/* order to be printed (-1=none, 0=active) */
-long	totsize;	/* total print job size in bytes */
-int	first;		/* first file in ``files'' column? */
-int	col;		/* column on screen */
-char	file[132];	/* print file name */
+static int	termwidth;
+static int	col;		/* column on screen */
+static char	current[NAME_MAX]; /* current file being printed */
+static char	file[NAME_MAX];	/* print file name */
+static int	first;		/* first file in ``files'' column? */
+static int	lflag;		/* long output option */
+static off_t	totsize;	/* total print job size in bytes */
 
-char	*head0 = "Rank   Owner      Job  Files";
-char	*head1 = "Total Size\n";
+static const char *head0 = "Rank   Owner      Job  Files";
+static const char *head1 = "Total Size\n";
+
+static void	alarmer(int);
+static void	inform(char *, int);
 
 /*
  * Display the current state of the queue. Format = 1 if long format.
  */
-displayq(format)
-	int format;
+void
+displayq(int format)
 {
-	register struct queue *q;
-	register int i, nitems, fd;
-	register char	*cp;
+	struct queue *q;
+	int i, rank, nitems, fd, ret, len;
+	char *cp, *ecp, *p;
 	struct queue **queue;
+	struct winsize win;
 	struct stat statb;
 	FILE *fp;
-	char c;
+
+	termwidth = 80;
+	if (isatty(STDOUT_FILENO)) {
+		if ((p = getenv("COLUMNS")) != NULL)
+			termwidth = atoi(p);
+		else if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &win) == 0 &&
+		    win.ws_col > 0)
+			termwidth = win.ws_col;
+	}
+	if (termwidth < 60)
+		termwidth = 60;
 
 	lflag = format;
 	totsize = 0;
-	rank = -1;
-
-	if ((i = pgetent(line, printer)) < 0)
-		fatal("cannot open printer description file");
-	else if (i == 0)
+	if ((i = cgetent(&bp, printcapdb, printer)) == -2)
+		fatal("can't open printer description file");
+	else if (i == -1)
 		fatal("unknown printer");
-	if ((LP = pgetstr("lp", &bp)) == NULL)
+	else if (i == -3)
+		fatal("potential reference loop detected in printcap file");
+	if (cgetstr(bp, DEFLP, &LP) < 0)
 		LP = _PATH_DEFDEVLP;
-	if ((RP = pgetstr("rp", &bp)) == NULL)
+	if (cgetstr(bp, "rp", &RP) < 0)
 		RP = DEFLP;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) < 0)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp,"lo", &LO) < 0)
 		LO = DEFLOCK;
-	if ((ST = pgetstr("st", &bp)) == NULL)
+	if (cgetstr(bp, "st", &ST) < 0)
 		ST = DEFSTAT;
-	RM = pgetstr("rm", &bp);
-	if (cp = checkremote())
+	cgetstr(bp, "rm", &RM);
+	if ((cp = checkremote()) != NULL)
 		printf("Warning: %s\n", cp);
 
 	/*
 	 * Print out local queue
 	 * Find all the control files in the spooling directory
 	 */
+	PRIV_START;
 	if (chdir(SD) < 0)
 		fatal("cannot chdir to spooling directory");
+	PRIV_END;
 	if ((nitems = getq(&queue)) < 0)
 		fatal("cannot examine spooling area\n");
-	if (stat(LO, &statb) >= 0) {
-		if (statb.st_mode & 0100) {
-			if (sendtorem)
+	PRIV_START;
+	ret = stat(LO, &statb);
+	PRIV_END;
+	if (ret >= 0) {
+		if (statb.st_mode & S_IXUSR) {
+			if (remote)
 				printf("%s: ", host);
 			printf("Warning: %s is down: ", printer);
-			fd = open(ST, O_RDONLY);
-			if (fd >= 0) {
-				(void) flock(fd, LOCK_SH);
+			PRIV_START;
+			fd = safe_open(ST, O_RDONLY|O_NOFOLLOW, 0);
+			PRIV_END;
+			if (fd >= 0 && flock(fd, LOCK_SH) == 0) {
 				while ((i = read(fd, line, sizeof(line))) > 0)
-					(void) fwrite(line, 1, i, stdout);
-				(void) close(fd);	/* unlocks as well */
+					(void)fwrite(line, 1, i, stdout);
+				(void)close(fd);	/* unlocks as well */
 			} else
 				putchar('\n');
 		}
-		if (statb.st_mode & 010) {
-			if (sendtorem)
+		if (statb.st_mode & S_IXGRP) {
+			if (remote)
 				printf("%s: ", host);
 			printf("Warning: %s queue is turned off\n", printer);
 		}
 	}
 
 	if (nitems) {
-		fp = fopen(LO, "r");
-		if (fp == NULL)
-			warn();
-		else {
+		PRIV_START;
+		fd = safe_open(LO, O_RDONLY|O_NOFOLLOW, 0);
+		PRIV_END;
+		if (fd < 0 || (fp = fdopen(fd, "r")) == NULL) {
+			if (fd >= 0)
+				close(fd);
+			nodaemon();
+		} else {
 			/* get daemon pid */
 			cp = current;
-			while ((*cp = getc(fp)) != EOF && *cp != '\n')
-				cp++;
+			ecp = cp + sizeof(current) - 1;
+			while ((i = getc(fp)) != EOF && i != '\n') {
+				if (cp < ecp)
+					*cp++ = i;
+			}
 			*cp = '\0';
 			i = atoi(current);
-			if (i <= 0 || kill(i, 0) < 0)
-				warn();
-			else {
+			if (i <= 0) {
+				ret = -1;
+			} else {
+				PRIV_START;
+				ret = kill(i, 0);
+				PRIV_END;
+			}
+			if (ret < 0 && errno != EPERM) {
+				nodaemon();
+			} else {
 				/* read current file name */
 				cp = current;
-				while ((*cp = getc(fp)) != EOF && *cp != '\n')
-					cp++;
+		    		ecp = cp + sizeof(current) - 1;
+				while ((i = getc(fp)) != EOF && i != '\n') {
+					if (cp < ecp)
+						*cp++ = i;
+				}
 				*cp = '\0';
 				/*
 				 * Print the status file.
 				 */
-				if (sendtorem)
+				if (remote)
 					printf("%s: ", host);
-				fd = open(ST, O_RDONLY);
-				if (fd >= 0) {
-					(void) flock(fd, LOCK_SH);
+				PRIV_START;
+				fd = safe_open(ST, O_RDONLY|O_NOFOLLOW, 0);
+				PRIV_END;
+				if (fd >= 0 && flock(fd, LOCK_SH) == 0) {
 					while ((i = read(fd, line, sizeof(line))) > 0)
-						(void) fwrite(line, 1, i, stdout);
-					(void) close(fd);	/* unlocks as well */
+						(void)fwrite(line, 1, i, stdout);
+					(void)close(fd);	/* unlocks as well */
 				} else
 					putchar('\n');
 			}
-			(void) fclose(fp);
+			(void)fclose(fp);
 		}
 		/*
 		 * Now, examine the control files and print out the jobs to
@@ -173,14 +228,18 @@ displayq(format)
 		 */
 		if (!lflag)
 			header();
-		for (i = 0; i < nitems; i++) {
+		/* The currently printed job is treated specially. */
+		if (!remote && current[0] != '\0')
+			inform(current, 0);
+		for (i = 0, rank = 1; i < nitems; i++) {
 			q = queue[i];
-			inform(q->q_name);
+			if (remote || strcmp(current, q->q_name) != 0)
+				inform(q->q_name, rank++);
 			free(q);
 		}
 		free(queue);
 	}
-	if (!sendtorem) {
+	if (!remote) {
 		if (nitems == 0)
 			puts("no entries");
 		return;
@@ -192,49 +251,74 @@ displayq(format)
 	 */
 	if (nitems)
 		putchar('\n');
-	(void) snprintf(line, sizeof(line), "%c%s", format + '\3', RP);
+	(void)snprintf(line, sizeof(line), "%c%s", format + '\3', RP);
 	cp = line;
-		cp += strlen(cp);
-	for (i = 0; i < requests; i++) {
-		char numbuf[32];
-		snprintf(numbuf, 32, " %d\0", requ[i]);
-		if (cp + strlen(numbuf) > line + sizeof(line)-2) {
+	cp += strlen(cp);
+	for (i = 0; i < requests && cp - line < sizeof(line) - 1; i++) {
+		len = line + sizeof(line) - cp;
+		if (snprintf(cp, len, " %d", requ[i]) >= len) {
+			cp += strlen(cp);
 			break;
 		}
-		(void) strcpy(cp, numbuf);
-		cp += strlen(numbuf);
-	}
-	for (i = 0; i < users; i++) {
-		if (cp + strlen(user[i]) > line + sizeof(line)-3) {
-			break;
-		}
-		*cp++ = ' ';
-		(void) strcpy(cp, user[i]);
 		cp += strlen(cp);
 	}
-	strcat(line, "\n");
-	fd = getport(RM);
+	for (i = 0; i < users && cp - line < sizeof(line) - 1; i++) {
+		len = line + sizeof(line) - cp;
+		if (snprintf(cp, len, " %s", user[i]) >= len) {
+			cp += strlen(cp);
+			break;
+		}
+	}
+	if (cp-line < sizeof(line) - 1)
+		strlcat(line, "\n", sizeof(line));
+	else
+		line[sizeof(line) - 2] = '\n';
+	fd = getport(RM, 0);
 	if (fd < 0) {
 		if (from != host)
 			printf("%s: ", host);
-		printf("connection to %s is down\n", RM);
+		(void)printf("connection to %s is down\n", RM);
 	}
 	else {
+		struct sigaction osa, nsa;
+		char *visline;
+
 		i = strlen(line);
 		if (write(fd, line, i) != i)
 			fatal("Lost connection");
-		while ((i = read(fd, line, sizeof(line))) > 0)
-			(void) fwrite(line, 1, i, stdout);
-		(void) close(fd);
+		memset(&nsa, 0, sizeof(nsa));
+		nsa.sa_handler = alarmer;
+		sigemptyset(&nsa.sa_mask);
+		nsa.sa_flags = 0;
+		(void)sigaction(SIGALRM, &nsa, &osa);
+		alarm(wait_time);
+		if ((visline = (char *)malloc(4 * sizeof(line) + 1)) == NULL)
+			fatal("Out of memory");
+		while ((i = read(fd, line, sizeof(line))) > 0) {
+			i = strvisx(visline, line, i, VIS_SAFE|VIS_NOSLASH);
+			(void)fwrite(visline, 1, i, stdout);
+			alarm(wait_time);
+		}
+		alarm(0);
+		(void)sigaction(SIGALRM, &osa, NULL);
+		free(visline);
+		(void)close(fd);
 	}
 }
 
+static void
+alarmer(int s)
+{
+	/* nothing */
+}
+
 /*
  * Print a warning message if there is no daemon present.
  */
-warn()
+void
+nodaemon(void)
 {
-	if (sendtorem)
+	if (remote)
 		printf("\n%s: ", host);
 	puts("Warning: no daemon present");
 	current[0] = '\0';
@@ -243,32 +327,34 @@ warn()
 /*
  * Print the header for the short listing format
  */
-header()
+void
+header(void)
 {
 	printf(head0);
 	col = strlen(head0)+1;
-	blankfill(SIZCOL);
+	blankfill(termwidth - (80 - SIZCOL));
 	printf(head1);
 }
 
-inform(cf)
-	char *cf;
+static void
+inform(char *cf, int rank)
 {
-	register int j, k;
-	register char *cp;
-	FILE *cfp;
+	int fd, j;
+	FILE *cfp = NULL;
 
 	/*
 	 * There's a chance the control file has gone away
 	 * in the meantime; if this is the case just keep going
 	 */
-	if ((cfp = fopen(cf, "r")) == NULL)
+	PRIV_START;
+	fd = safe_open(cf, O_RDONLY|O_NOFOLLOW, 0);
+	PRIV_END;
+	if (fd < 0 || (cfp = fdopen(fd, "r")) == NULL) {
+		if (fd >= 0)
+			close(fd);
 		return;
+	}
 
-	if (rank < 0)
-		rank = 0;
-	if (sendtorem || garbage || strcmp(cf, current))
-		rank++;
 	j = 0;
 	while (getline(cfp)) {
 		switch (line[0]) {
@@ -295,10 +381,8 @@ inform(cf)
 		default: /* some format specifer and file name? */
 			if (line[0] < 'a' || line[0] > 'z')
 				continue;
-			if (j == 0 || strcmp(file, line+1) != 0) {
-				(void) strncpy(file, line+1, sizeof(file)-1);
-				file[sizeof(file)-1] = '\0';
-			}
+			if (j == 0 || strcmp(file, line+1) != 0)
+				(void)strlcpy(file, line+1, sizeof(file));
 			j++;
 			continue;
 		case 'N':
@@ -309,17 +393,17 @@ inform(cf)
 	}
 	fclose(cfp);
 	if (!lflag) {
-		blankfill(SIZCOL);
-		printf("%ld bytes\n", totsize);
+		blankfill(termwidth - (80 - SIZCOL));
+		printf("%lld bytes\n", (long long)totsize);
 		totsize = 0;
 	}
 }
 
-inlist(name, file)
-	char *name, *file;
+int
+inlist(char *name, char *file)
 {
-	register int *r, n;
-	register char **u, *cp;
+	int *r, n;
+	char **u, *cp;
 
 	if (users == 0 && requests == 0)
 		return(1);
@@ -340,8 +424,8 @@ inlist(name, file)
 	return(0);
 }
 
-show(nfile, file, copies)
-	register char *nfile, *file;
+void
+show(char *nfile, char *file, int copies)
 {
 	if (strcmp(nfile, " ") == 0)
 		nfile = "(standard input)";
@@ -354,8 +438,8 @@ show(nfile, file, copies)
 /*
  * Fill the line with blanks to the specified column
  */
-blankfill(n)
-	register int n;
+void
+blankfill(int n)
 {
 	while (col++ < n)
 		putchar(' ');
@@ -364,10 +448,10 @@ blankfill(n)
 /*
  * Give the abbreviated dump of the file names
  */
-dump(nfile, file, copies)
-	char *nfile, *file;
+void
+dump(char *nfile, char *file, int copies)
 {
-	register short n, fill;
+	int n, fill;
 	struct stat lbuf;
 
 	/*
@@ -375,10 +459,11 @@ dump(nfile, file, copies)
 	 *  (leaving room for the total size)
 	 */
 	 fill = first ? 0 : 2;	/* fill space for ``, '' */
-	 if (((n = strlen(nfile)) + col + fill) >= SIZCOL-4) {
-		if (col < SIZCOL) {
+	 if (((n = strlen(nfile)) + col + fill) >=
+	     (termwidth - (80 - SIZCOL)) - 4) {
+		if (col < (termwidth - (80 - SIZCOL))) {
 			printf(" ..."), col += 4;
-			blankfill(SIZCOL);
+			blankfill(termwidth - (80 - SIZCOL));
 		}
 	} else {
 		if (first)
@@ -388,15 +473,17 @@ dump(nfile, file, copies)
 		printf("%s", nfile);
 		col += n+fill;
 	}
+	PRIV_START;
 	if (*file && !stat(file, &lbuf))
 		totsize += copies * lbuf.st_size;
+	PRIV_END;
 }
 
 /*
  * Print the long info about the file
  */
-ldump(nfile, file, copies)
-	char *nfile, *file;
+void
+ldump(char *nfile, char *file, int copies)
 {
 	struct stat lbuf;
 
@@ -406,7 +493,7 @@ ldump(nfile, file, copies)
 	else
 		printf("%-32s", nfile);
 	if (*file && !stat(file, &lbuf))
-		printf(" %ld bytes", lbuf.st_size);
+		printf(" %lld bytes", (long long)lbuf.st_size);
 	else
 		printf(" ??? bytes");
 	putchar('\n');
@@ -416,9 +503,10 @@ ldump(nfile, file, copies)
  * Print the job's rank in the queue,
  *   update col for screen management
  */
-prank(n)
+void
+prank(int n)
 {
-	char line[100];
+	char rline[100];
 	static char *r[] = {
 		"th", "st", "nd", "rd", "th", "th", "th", "th", "th", "th"
 	};
@@ -429,9 +517,9 @@ prank(n)
 		return;
 	}
 	if ((n/10)%10 == 1)
-		(void) snprintf(line, 100, "%dth", n);
+		(void)snprintf(rline, sizeof(rline), "%dth", n);
 	else
-		(void) snprintf(line, 100, "%d%s", n, r[n%10]);
-	col += strlen(line);
-	printf("%s", line);
+		(void)snprintf(rline, sizeof(rline), "%d%s", n, r[n%10]);
+	col += strlen(rline);
+	printf("%s", rline);
 }
