@@ -1,4 +1,5 @@
-/*	$OpenBSD: lpf.c,v 1.4 1997/07/19 07:11:42 deraadt Exp $	*/
+/*	$OpenBSD: lpf.c,v 1.10 2003/06/02 23:36:53 millert Exp $	*/
+/*	$NetBSD: lpf.c,v 1.8 2000/04/29 00:12:32 abs Exp $	*/
 
 /*
  * Copyright (c) 1983, 1993
@@ -12,11 +13,7 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. All advertising materials mentioning features or use of this software
- *    must display the following acknowledgement:
- *	This product includes software developed by the University of
- *	California, Berkeley and its contributors.
- * 4. Neither the name of the University nor the names of its contributors
+ * 3. Neither the name of the University nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  *
@@ -34,16 +31,16 @@
  */
 
 #ifndef lint
-static char copyright[] =
+static const char copyright[] =
 "@(#) Copyright (c) 1983, 1993\n\
 	The Regents of the University of California.  All rights reserved.\n";
 #endif /* not lint */
 
 #ifndef lint
 #if 0
-static char sccsid[] = "@(#)lpf.c	8.1 (Berkeley) 6/6/93";
+static const char sccsid[] = "@(#)lpf.c	8.1 (Berkeley) 6/6/93";
 #else
-static char rcsid[] = "$OpenBSD: lpf.c,v 1.4 1997/07/19 07:11:42 deraadt Exp $";
+static const char rcsid[] = "$OpenBSD: lpf.c,v 1.10 2003/06/02 23:36:53 millert Exp $";
 #endif
 #endif /* not lint */
 
@@ -52,14 +49,13 @@ static char rcsid[] = "$OpenBSD: lpf.c,v 1.4 1997/07/19 07:11:42 deraadt Exp $";
  *	with ^H's to overwritten lines.  Thus this works like 'ul'
  *	but is much better: it can handle more than 2 overwrites
  *	and it is written with some style.
- *	modified by kls to use register references instead of arrays
- *	to try to gain a little speed.
  */
 
 #include <signal.h>
-#include <unistd.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <string.h>
+#include <unistd.h>
 
 #define MAXWIDTH  132
 #define MAXREP    10
@@ -77,56 +73,51 @@ char	*name;		/* user's login name */
 char	*host;		/* user's machine name */
 char	*acctfile;	/* accounting information file */
 
+__dead void usage(void);
+
 int
-main(argc, argv) 
-	int argc;
-	char *argv[];
+main(int argc, char **argv) 
 {
-	register FILE *p = stdin, *o = stdout;
-	register int i, col;
-	register char *cp;
-	int done, linedone, maxrep;
-	char ch, *limit;
+	FILE *p = stdin, *o = stdout;
+	int i, col;
+	char *cp;
+	int done, linedone, maxrep, ch;
+	char *limit;
 
-	while (--argc) {
-		if (*(cp = *++argv) == '-') {
-			switch (cp[1]) {
-			case 'n':
-				argc--;
-				name = *++argv;
-				break;
-
-			case 'h':
-				argc--;
-				host = *++argv;
-				break;
-
-			case 'w':
-				if ((i = atoi(&cp[2])) > 0 && i <= MAXWIDTH)
-					width = i;
-				break;
-
-			case 'l':
-				length = atoi(&cp[2]);
-				break;
-
-			case 'i':
-				indent = atoi(&cp[2]);
-				break;
-
-			case 'r':	/* map nl->cr-nl */
-				onlcr++;
-				break;
-
-			case 'c':	/* Print control chars */
-				literal++;
-				break;
-			}
-		} else
-			acctfile = cp;
+	while ((ch = getopt(argc, argv, "crh:i:l:n:w:")) != -1) {
+		switch (ch) {
+		case 'n':
+			name = optarg;
+			break;
+		case 'h':
+			host = optarg;
+			break;
+		case 'w':
+			if ((i = atoi(optarg)) > 0 && i <= MAXWIDTH)
+				width = i;
+			break;
+		case 'l':
+			length = atoi(optarg);
+			break;
+		case 'i':
+			indent = atoi(optarg);
+			break;
+		case 'r':	/* map nl->cr-nl */
+			onlcr++;
+			break;
+		case 'c':	/* Print control chars */
+			literal++;
+			break;
+		default:
+			usage();
+		}
 	}
+	argc -= optind;
+	argv += optind;
+	if (argc)
+		acctfile = *argv;
 
-	for (cp = buf[0], limit = buf[MAXREP]; cp < limit; *cp++ = ' ');
+	memset(buf, ' ',  sizeof(buf));
 	done = 0;
 	
 	while (!done) {
@@ -177,7 +168,7 @@ main(argc, argv)
 				}
 
 			default:
-				if ((col >= width) || (!literal && ch < ' ')) {
+				if (col >= width || (!literal && ch < ' ')) {
 					col++;
 					break;
 				}
@@ -229,3 +220,13 @@ main(argc, argv)
 	}
 	exit(0);
 }
+
+__dead void
+usage(void)
+{
+	extern char *__progname;
+
+	fprintf(stderr, "usage: %s [-c] [-r] [-h host] [-i indent] [-l length]"
+	    " [-n name] [-w width] [acctfile]\n", __progname);
+	exit(1);
+}
