@@ -1,4 +1,5 @@
-/*	$OpenBSD: recvjob.c,v 1.12 1997/10/05 11:37:01 deraadt Exp $	*/
+/*	$OpenBSD: recvjob.c,v 1.23 2003/06/02 23:36:53 millert Exp $	*/
+/*	$NetBSD: recvjob.c,v 1.14 2001/12/04 22:52:44 christos Exp $	*/
 
 /*
  * Copyright (c) 1983, 1993
@@ -13,11 +14,7 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. All advertising materials mentioning features or use of this software
- *    must display the following acknowledgement:
- *	This product includes software developed by the University of
- *	California, Berkeley and its contributors.
- * 4. Neither the name of the University nor the names of its contributors
+ * 3. Neither the name of the University nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  *
@@ -35,16 +32,16 @@
  */
 
 #ifndef lint
-static char copyright[] =
+static const char copyright[] =
 "@(#) Copyright (c) 1983, 1993\n\
 	The Regents of the University of California.  All rights reserved.\n";
 #endif /* not lint */
 
 #ifndef lint
 #if 0
-static char sccsid[] = "@(#)recvjob.c	8.2 (Berkeley) 4/27/95";
+static const char sccsid[] = "@(#)recvjob.c	8.2 (Berkeley) 4/27/95";
 #else
-static char rcsid[] = "$OpenBSD: recvjob.c,v 1.12 1997/10/05 11:37:01 deraadt Exp $";
+static const char rcsid[] = "$OpenBSD: recvjob.c,v 1.23 2003/06/02 23:36:53 millert Exp $";
 #endif
 #endif /* not lint */
 
@@ -62,31 +59,34 @@ static char rcsid[] = "$OpenBSD: recvjob.c,v 1.12 1997/10/05 11:37:01 deraadt Ex
 #include <dirent.h>
 #include <syslog.h>
 #include <stdio.h>
+#include <errno.h>
 #include <stdlib.h>
 #include <string.h>
+#include <stdarg.h>
 #include "lp.h"
 #include "lp.local.h"
 #include "extern.h"
 #include "pathnames.h"
 
-#define ack()	(void) write(1, sp, 1);
+#define ack()	(void)write(STDOUT_FILENO, sp, 1);
 
 static char	 dfname[NAME_MAX];	/* data files */
 static int	 minfree;       /* keep at least minfree blocks available */
 static char	*sp = "";
 static char	 tfname[NAME_MAX];	/* tmp copy of cf before linking */
 
-static int        chksize __P((int));
-static void       frecverr __P((const char *, ...));
-static int        noresponse __P((void));
-static void       rcleanup __P((int));
-static int        read_number __P((char *));
-static int        readfile __P((char *, int));
-static int        readjob __P((void));
+static int        chksize(int);
+static void       frecverr(const char *, ...)
+	__attribute__((__format__(__printf__, 1, 2)));
+static int        noresponse(void);
+static void       rcleanup(int);
+static int        read_number(char *);
+static int        readfile(char *, int);
+static int        readjob(void);
 
 
 void
-recvjob()
+recvjob(void)
 {
 	struct stat stb;
 	int status;
@@ -108,11 +108,13 @@ recvjob()
 	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
 
-	(void) close(2);			/* set up log file */
+	(void)close(2);			/* set up log file */
+	PRIV_START;
 	if (open(LF, O_WRONLY|O_APPEND, 0664) < 0) {
 		syslog(LOG_ERR, "%s: %m", LF);
-		(void) open(_PATH_DEVNULL, O_WRONLY);
+		(void)open(_PATH_DEVNULL, O_WRONLY);
 	}
+	PRIV_END;
 
 	if (chdir(SD) < 0)
 		frecverr("%s: %s: %m", printer, SD);
@@ -124,6 +126,7 @@ recvjob()
 		}
 	} else if (stat(SD, &stb) < 0)
 		frecverr("%s: %s: %m", printer, SD);
+
 	minfree = 2 * read_number("minfree");	/* scale KB to 512 blocks */
 	signal(SIGTERM, rcleanup);
 	signal(SIGPIPE, rcleanup);
@@ -134,13 +137,13 @@ recvjob()
 
 /*
  * Read printer jobs sent by lpd and copy them to the spooling directory.
- * Return the number of jobs successfully transfered.
+ * Return the number of jobs successfully transferred.
  */
 static int
-readjob()
+readjob(void)
 {
-	register int size, nfiles;
-	register char *cp;
+	int size, nfiles;
+	char *cp;
 
 	ack();
 	nfiles = 0;
@@ -150,14 +153,14 @@ readjob()
 		 */
 		cp = line;
 		do {
-			if ((size = read(1, cp, 1)) != 1) {
+			if ((size = read(STDOUT_FILENO, cp, 1)) != 1) {
 				if (size < 0)
 					frecverr("%s: Lost connection",
 					    printer);
 				return(nfiles);
 			}
-		} while (*cp++ != '\n' && (cp - line + 1) < sizeof line);
-		if (cp - line + 1 >= sizeof line)
+		} while (*cp++ != '\n' && (cp - line + 1) < sizeof(line));
+		if (cp - line + 1 >= sizeof(line))
 			frecverr("readjob overflow");
 		*--cp = '\0';
 		cp = line;
@@ -178,25 +181,32 @@ readjob()
 			 * something different than what gethostbyaddr()
 			 * returns
 			 */
-			strncpy(cp + 6, from, sizeof(line) + line - cp - 7);
-			line[sizeof(line) -1 ] = '\0';
-			strncpy(tfname, cp, sizeof tfname-1);
-			tfname[sizeof tfname-1] = '\0';
+			strlcpy(cp + 6, from, sizeof(line) + line - cp - 6);
+			if (strchr(cp, '/'))
+				frecverr("readjob: %s: illegal path name", cp);
+			strlcpy(tfname, cp, sizeof(tfname));
 			tfname[0] = 't';
-			if (strchr(tfname, '/'))
-				frecverr("readjob: %s: illegal path name",
-				    tfname);
 			if (!chksize(size)) {
-				(void) write(1, "\2", 1);
+				(void)write(STDOUT_FILENO, "\2", 1);
 				continue;
 			}
+			/*
+			 * XXX
+			 * We blindly believe what the remote host puts
+			 * for the path to the df file.  In general this
+			 * is OK since we don't allow paths with '/' in
+			 * them.  Still, it would be better to sanity
+			 * check the cf file sent to us and make the
+			 * df name match the cf name we used.  That way
+			 * we avoid any possible collisions.
+			 */
 			if (!readfile(tfname, size)) {
 				rcleanup(0);
 				continue;
 			}
 			if (link(tfname, cp) < 0)
-				frecverr("%s: %m", tfname);
-			(void) unlink(tfname);
+				frecverr("link %s %s: %m", tfname, cp);
+			(void)unlink(tfname);
 			tfname[0] = '\0';
 			nfiles++;
 			continue;
@@ -207,16 +217,14 @@ readjob()
 				size = size * 10 + (*cp++ - '0');
 			if (*cp++ != ' ')
 				break;
+			if (strchr(cp, '/'))
+				frecverr("readjob: %s: illegal path name", cp);
 			if (!chksize(size)) {
-				(void) write(1, "\2", 1);
+				(void)write(STDOUT_FILENO, "\2", 1);
 				continue;
 			}
-			(void) strncpy(dfname, cp, sizeof dfname-1);
-			dfname[sizeof dfname-1] = '\0';
-			if (strchr(dfname, '/'))
-				frecverr("readjob: %s: illegal path name",
-					dfname);
-			(void) readfile(dfname, size);
+			(void)strlcpy(dfname, cp, sizeof(dfname));
+			(void)readfile(dfname, size);
 			continue;
 		}
 		frecverr("protocol screwup: %s", line);
@@ -227,17 +235,14 @@ readjob()
  * Read files send by lpd and copy them to the spooling directory.
  */
 static int
-readfile(file, size)
-	char *file;
-	int size;
+readfile(char *file, int size)
 {
-	register char *cp;
+	char *cp;
 	char buf[BUFSIZ];
-	register int i, j, amt;
+	int i, j, amt;
 	int fd, err;
 
-	fd = open(file, O_CREAT|O_EXCL|O_WRONLY, FILMOD);
-	if (fd < 0)
+	if ((fd = open(file, O_CREAT|O_EXCL|O_WRONLY, FILMOD)) < 0)
 		frecverr("readfile: %s: illegal path name: %m", file);
 	ack();
 	err = 0;
@@ -247,7 +252,7 @@ readfile(file, size)
 		if (i + amt > size)
 			amt = size - i;
 		do {
-			j = read(1, cp, amt);
+			j = read(STDOUT_FILENO, cp, amt);
 			if (j <= 0)
 				frecverr("Lost connection");
 			amt -= j;
@@ -261,12 +266,12 @@ readfile(file, size)
 			break;
 		}
 	}
-	(void) close(fd);
+	(void)close(fd);
 	if (err)
 		frecverr("%s: write error", file);
 	if (noresponse()) {		/* file sent had bad data in it */
 		if (strchr(file, '/') == NULL)
-			(void) unlink(file);
+			(void)unlink(file);
 		return(0);
 	}
 	ack();
@@ -274,11 +279,11 @@ readfile(file, size)
 }
 
 static int
-noresponse()
+noresponse(void)
 {
 	char resp;
 
-	if (read(1, &resp, 1) != 1)
+	if (read(STDOUT_FILENO, &resp, 1) != 1)
 		frecverr("Lost connection");
 	if (resp == '\0')
 		return(0);
@@ -290,8 +295,7 @@ noresponse()
  * 1 == OK, 0 == Not OK.
  */
 static int
-chksize(size)
-	int size;
+chksize(int size)
 {
 	int spacefree;
 	struct statfs sfb;
@@ -308,15 +312,14 @@ chksize(size)
 }
 
 static int
-read_number(fn)
-	char *fn;
+read_number(char *fn)
 {
 	char lin[80];
-	register FILE *fp;
+	FILE *fp;
 
 	if ((fp = fopen(fn, "r")) == NULL)
 		return (0);
-	if (fgets(lin, 80, fp) == NULL) {
+	if (fgets(lin, sizeof(lin), fp) == NULL) {
 		fclose(fp);
 		return (0);
 	}
@@ -325,47 +328,35 @@ read_number(fn)
 }
 
 /*
- * Remove all the files associated with the current job being transfered.
+ * Remove all the files associated with the current job being transferred.
  */
 static void
-rcleanup(signo)
-	int signo;
+rcleanup(int signo)
 {
+	int save_errno = errno;
+
 	if (tfname[0] && strchr(tfname, '/') == NULL)
-		(void) unlink(tfname);
+		(void)unlink(tfname);
 	if (dfname[0] && strchr(dfname, '/') == NULL) {
 		do {
 			do
-				(void) unlink(dfname);
-			while (dfname[2]-- != 'A');
+				(void)unlink(dfname);
+			while (dfname[2]-- != 'A')
+				;
 			dfname[2] = 'z';
 		} while (dfname[0]-- != 'd');
 	}
 	dfname[0] = '\0';
+	errno = save_errno;
 }
 
-#ifdef __STDC__
-#include <stdarg.h>
-#else
-#include <varargs.h>
-#endif
-
 static void
-#ifdef __STDC__
 frecverr(const char *msg, ...)
-#else
-frecverr(msg, va_alist)
-	char *msg;
-        va_dcl
-#endif
 {
 	extern char fromb[];
 	va_list ap;
-#ifdef __STDC__
+
 	va_start(ap, msg);
-#else
-	va_start(ap);
-#endif
 	rcleanup(0);
 	syslog(LOG_ERR, "%s", fromb);
 	vsyslog(LOG_ERR, msg, ap);
